#include "print_log.h"
#include "usart.h"
#include <stdarg.h>
#include <string.h>
#include <stdio.h>

uint8_t transfer_enable = 0;    // 标记串口的发送dma是否正在工作

#define QUEUE_MAX   4096
#define print_queue_buff_SIZE  128
#define DMA_BUFFER_SIZE 512

uint8_t print_queue[QUEUE_MAX];
uint8_t print_queue_buff[print_queue_buff_SIZE];

uint16_t queue_write_point = 1000;
uint16_t queue_read_point = 1000;

uint8_t print_dma_buff[DMA_BUFFER_SIZE];

uint8_t print_enable = 0;

void print_log(char *fmt,...)
{
    uint16_t array_size = 0;
    va_list arg;
    va_start(arg,fmt);
    vsprintf((char*)print_queue_buff,fmt,arg);
    va_end(arg);

    // 把格式化之后的内容放入打印队列
    array_size = strlen((const char *)print_queue_buff);
    print_str(print_queue_buff, array_size);
}

void print_str(uint8_t* str, uint16_t len)
{
    if(queue_write_point + len < QUEUE_MAX)
    {
        memcpy(&print_queue[queue_write_point],str,len);
        queue_write_point += len;
    }
    else
    {
        memcpy(&print_queue[queue_write_point],str,QUEUE_MAX - queue_write_point);
        memcpy(&print_queue[0],&str[QUEUE_MAX - queue_write_point],len + queue_write_point - QUEUE_MAX);
        queue_write_point = len + queue_write_point - QUEUE_MAX;
    }
    if(transfer_enable == 0)
    {
        transfer_enable = 1;
        start_dma_transfer();
    }
}

uint16_t dma_length;
void start_dma_transfer(void)
{
    if(queue_read_point != queue_write_point) // 确保不是误操作
    {
        if(queue_read_point < queue_write_point)    // 头在前，尾在后面
        {
            if(queue_read_point + DMA_BUFFER_SIZE <= queue_write_point)
            {
                memcpy(print_dma_buff,&print_queue[queue_read_point],DMA_BUFFER_SIZE);
                dma_length = DMA_BUFFER_SIZE;
                queue_read_point += DMA_BUFFER_SIZE;
            }
            else
            {
                memcpy(print_dma_buff,&print_queue[queue_read_point],queue_write_point - queue_read_point);
                dma_length = queue_write_point - queue_read_point;
                queue_read_point = queue_write_point;
            }
        }
        else        // 头在后，尾在前
        {
            if(queue_read_point + DMA_BUFFER_SIZE <= QUEUE_MAX)
            {
                memcpy(print_dma_buff,&print_queue[queue_read_point],DMA_BUFFER_SIZE);
                dma_length = DMA_BUFFER_SIZE;
                queue_read_point += DMA_BUFFER_SIZE;
            }
            else
            {
                memcpy(print_dma_buff,&print_queue[queue_read_point],QUEUE_MAX - queue_read_point);
                dma_length = QUEUE_MAX - queue_read_point;
                queue_read_point = 0;
            }
        }
        HAL_UART_Transmit_DMA(&PRINT_USART, print_dma_buff, dma_length);
    }
}

void HAL_UART_TxCpltCallback(UART_HandleTypeDef *huart)
{
    if(huart == &PRINT_USART) // 处理串口2的发送DMA
    {
        if(queue_read_point != queue_write_point)
        {
            start_dma_transfer();   // 开始传输
        }
        else
        {
            transfer_enable = 0;    // 结束传输，标记
        }
    }
}
